﻿/*
	Copyright (c) 2017 TOSHIBA Digital Solutions Corporation

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU Affero General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Affero General Public License for more details.

	You should have received a copy of the GNU Affero General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*!
	@file
	@brief Definition of SQL container
*/

#ifndef SQL_UTILS_CONTAINER_IMPL_H_
#define SQL_UTILS_CONTAINER_IMPL_H_

#include "sql_utils_container.h"

#include "transaction_service.h"
#include "query_processor.h"
#include "result_set.h"
#include "meta_store.h"

#include "sql_expression_core.h"

class BoolExpr;
class Expr;


struct SQLContainerImpl {

	typedef SQLType::Id ExpressionType;
	typedef SQLType::IndexType IndexType;

	typedef util::Vector<TupleColumnType> ColumnTypeList;
	typedef TupleValue::VarContext VarContext;

	typedef ::ColumnType ContainerColumnType;

	typedef BtreeMap::SearchContext BtreeSearchContext;
	typedef HashMap::SearchContext HashSearchContext;

	typedef BaseContainer::RowArray RowArray;
	typedef BaseContainer::RowArrayType RowArrayType;
	typedef RowArray::Column ContainerColumn;

	typedef SQLValues::ValueUtils ValueUtils;

	typedef SQLExprs::SyntaxExpr SyntaxExpr;
	typedef SQLExprs::SyntaxExprRefList SyntaxExprRefList;
	typedef SQLExprs::Expression Expression;

	typedef SQLExprs::IndexSelector IndexSelector;
	typedef SQLOps::ContainerLocation ContainerLocation;

	typedef SQLOps::TupleColumnList TupleColumnList;
	typedef SQLOps::OpContext OpContext;
	typedef SQLOps::Projection Projection;

	typedef SQLContainerUtils::ScanCursor::LatchTarget LatchTarget;

	class CursorImpl;

	struct Constants;

	struct ColumnCode;
	template<RowArrayType RAType, typename T, typename S, bool Nullable>
	class ColumnEvaluator;

	class ColumnExpression;
	class IdExpression;
	class ComparisonExpressionBase;
	template<typename Func> class ComparisonExpression;

	class GeneralCondEvaluator;
	template<typename Func, size_t V> class ComparisonEvaluator;
	class EmptyCondEvaluator;

	class GeneralProjector;
	class CountProjector;

	class CursorRef;

	template<RowArrayType RAType>
	struct ScannerHandler;
	class ScannerHandlerBase;
	class ScannerHandlerFactory;

	class BaseRegistrar;
	class DefaultRegistrar;

	struct SearchResult;
	class CursorScope;
	class MetaScanHandler;
	class TupleUpdateRowIdHandler;

	struct ContainerValueUtils;
	struct TQLTool;

	typedef void (*UpdatorFunc)(OpContext&, void*);
	typedef std::pair<UpdatorFunc, void*> UpdatorEntry;
};

class SQLContainerImpl::CursorImpl : public SQLContainerUtils::ScanCursor {
public:
	explicit CursorImpl(const Source &source);
	virtual ~CursorImpl();

	virtual void unlatch() throw();
	virtual void close() throw();
	void destroy() throw();

	virtual bool scanFull(OpContext &cxt, const Projection &proj);
	virtual bool scanIndex(
			OpContext &cxt, const Projection &proj,
			const SQLExprs::IndexConditionList &condList);
	virtual bool scanUpdates(OpContext &cxt, const Projection &proj);
	virtual bool scanMeta(
			OpContext &cxt, const Projection &proj, const Expression *pred);

	virtual void getIndexSpec(
			OpContext &cxt, SQLExprs::IndexSelector &selector);
	virtual bool isIndexLost();

	virtual uint32_t getOutputIndex();
	virtual void setOutputIndex(uint32_t index);

	void startScope(OpContext &cxt);
	void finishScope() throw();
	CursorScope& resolveScope();

	ContainerId getNextContainerId() const;
	void setNextContainerId(ContainerId containerId);

	BaseContainer& resolveContainer();
	BaseContainer* getContainer();

	TransactionContext& resolveTransactionContext();
	ResultSet& resolveResultSet();

	template<BaseContainer::RowArrayType RAType>
	void setRow(
			typename BaseContainer::RowArrayImpl<
					BaseContainer, RAType> &rowArrayImpl);

	bool isValueNull(const ContainerColumn &column);

	template<RowArrayType RAType, typename Ret, TupleColumnType T>
	Ret getValueDirect(
			SQLValues::ValueContext &cxt, const ContainerColumn &column);

	template<RowArrayType RAType>
	void writeValue(
			SQLValues::ValueContext &cxt, TupleList::Writer &writer,
			const TupleList::Column &destColumn,
			const ContainerColumn &srcColumn, bool forId);

	template<TupleColumnType T>
	static void writeValueAs(
			TupleList::Writer &writer, const TupleList::Column &column,
			const typename SQLValues::TypeUtils::Traits<T>::ValueType &value);

	ContainerRowScanner* tryCreateScanner(
			OpContext &cxt, const Projection &proj,
			const BaseContainer &container);
	void updateScanner(OpContext &cxt, bool finished);

	void clearResultSetPreserving(ResultSet &resultSet);
	void activateResultSetPreserving(ResultSet &resultSet);

	static void checkInputInfo(OpContext &cxt, const BaseContainer &container);

	static TupleColumnType resolveColumnType(
			const ColumnInfo &columnInfo, bool used, bool withStats);
	static ContainerColumn resolveColumn(
			BaseContainer &container, bool forId, uint32_t pos);
	bool isColumnNonNullable(uint32_t pos);

	void preparePartialSearch(
			OpContext &cxt, BtreeSearchContext &sc,
			const SQLExprs::IndexConditionList *targetCondList);
	bool finishPartialSearch(OpContext &cxt, BtreeSearchContext &sc);

	static void setUpSearchContext(
			OpContext &cxt, TransactionContext &txn,
			const SQLExprs::IndexConditionList &targetCondList,
			BtreeSearchContext &sc, const BaseContainer &container);

	IndexType acceptIndexCond(
			OpContext &cxt, const SQLExprs::IndexConditionList &targetCondList,
			util::Vector<uint32_t> &indexColumnList);

	void acceptSearchResult(
			OpContext &cxt, const SearchResult &result,
			const SQLExprs::IndexConditionList *targetCondList,
			ContainerRowScanner &scanner);

	static void acceptIndexInfo(
			OpContext &cxt, TransactionContext &txn, BaseContainer &container,
			const SQLExprs::IndexConditionList &targetCondList,
			const IndexInfo &info);

	void setIndexLost(OpContext &cxt);

	template<typename Container>
	static void scanRow(
			TransactionContext &txn, Container &container,
			const SearchResult &result, ContainerRowScanner &scanner);

	void initializeScanRange(
			TransactionContext &txn, BaseContainer &container);

	static void assignIndexInfo(
			TransactionContext &txn, BaseContainer &container,
			IndexSelector &indexSelector, const TupleColumnList &coumnList);
	static void getIndexInfoList(
			TransactionContext &txn, BaseContainer &container,
			util::Vector<IndexInfo> &indexInfoList);

	static TransactionContext& prepareTransactionContext(OpContext &cxt);

	static FullContainerKey* predicateToContainerKey(
			SQLValues::ValueContext &cxt, TransactionContext &txn,
			DataStore &dataStore, const Expression *pred,
			const ContainerLocation &location, util::String &dbNameStr);

	static const MetaContainerInfo& resolveMetaContainerInfo(ContainerId id);

	static DataStore* getDataStore(OpContext &cxt);
	static EventContext* getEventContext(OpContext &cxt);
	static TransactionManager* getTransactionManager(OpContext &cxt);
	static TransactionService* getTransactionService(OpContext &cxt);
	static PartitionTable* getPartitionTable(OpContext &cxt);
	static ClusterService* getClusterService(OpContext &cxt);
	static SQLExecutionManager* getExecutionManager(OpContext &cxt);
	static SQLService* getSQLService(OpContext &cxt);
	static const ResourceSet* getResourceSet(OpContext &cxt);

private:
	friend class CursorScope;

	typedef BaseContainer::RowArrayImpl<
			BaseContainer, BaseContainer::ROW_ARRAY_GENERAL> GeneralRowArray;
	typedef BaseContainer::RowArrayImpl<
			BaseContainer, BaseContainer::ROW_ARRAY_PLAIN> PlainRowArray;

	typedef util::Vector<UpdatorEntry> UpdatorList;

	template<
			TupleColumnType T,
			bool VarSize = SQLValues::TypeUtils::Traits<T>::SIZE_VAR>
	struct DirectValueAccessor;

	CursorImpl(const CursorImpl&);
	CursorImpl& operator=(const CursorImpl&);

	template<RowArrayType RAType, typename T>
	const T& getFixedValue(const ContainerColumn &column);

	template<RowArrayType RAType>
	const void* getFixedValueAddr(const ContainerColumn &column);

	template<RowArrayType RAType>
	const void* getVariableArray(const ContainerColumn &column);

	void initializeRowArray(TransactionContext &txn, BaseContainer &container);

	void destroyRowArray();

	RowArray* getRowArray();

	template<RowArrayType RAType>
	typename BaseContainer::RowArrayImpl<
			BaseContainer, RAType>* getRowArrayImpl(bool loaded);

	template<RowArrayType RAType>
	void setUpScannerHandler(
			OpContext &cxt, const Projection &proj,
			ContainerRowScanner::HandlerSet &handlerSet);

	void destroyUpdateRowIdHandler() throw();

	static RowId getMaxRowId(
			TransactionContext& txn, BaseContainer &container);

	static bool isColumnSchemaNullable(const ColumnInfo &info);

	bool updateNullsStats(const BaseContainer &container);
	static void getNullsStats(
			const BaseContainer &container, util::XArray<uint8_t> &nullsStats);

	static void initializeBits(util::XArray<uint8_t> &bits, size_t bitCount);
	static void setBit(util::XArray<uint8_t> &bits, size_t index, bool value);
	static bool getBit(const util::XArray<uint8_t> &bits, size_t index);

	ContainerLocation location_;
	uint32_t outIndex_;

	bool closed_;
	bool resultSetPreserving_;
	bool containerExpired_;
	bool indexLost_;
	bool resultSetLost_;
	bool nullsStatsChanged_;

	ResultSetId resultSetId_;
	ResultSetHolder resultSetHolder_;

	PartitionId lastPartitionId_;
	DataStore *lastDataStore_;
	BaseContainer *container_;
	ContainerType containerType_;
	SchemaVersionId schemaVersionId_;
	ContainerId nextContainerId_;
	RowId maxRowId_;

	util::LocalUniquePtr<RowArray> rowArray_;
	GeneralRowArray *generalRowArray_;
	PlainRowArray *plainRowArray_;
	uint8_t *row_;
	util::LocalUniquePtr<BaseObject> frontFieldCache_;

	util::AllocUniquePtr<CursorScope> scope_;
	LatchTarget latchTarget_;
	util::AllocUniquePtr<TupleUpdateRowIdHandler> updateRowIdHandler_;

	ContainerRowScanner *scanner_;
	util::XArray<uint8_t> initialNullsStats_;
	util::XArray<uint8_t> lastNullsStats_;
	int64_t indexLimit_;
	int64_t totalSearchCount_;
public:
	util::LocalUniquePtr<UpdatorList> updatorList_;
};

template<TupleColumnType T, bool VarSize>
struct SQLContainerImpl::CursorImpl::DirectValueAccessor {
	template<RowArrayType RAType, typename Ret>
	Ret access(
			SQLValues::ValueContext &cxt, CursorImpl &cursor,
			const ContainerColumn &column) const;
};

struct SQLContainerImpl::Constants {
	static const uint16_t SCHEMA_INVALID_COLUMN_ID;
	static const uint32_t ESTIMATED_ROW_ARRAY_ROW_COUNT;
};

struct SQLContainerImpl::ColumnCode {
	typedef SQLExprs::ExprFactoryContext ExprFactoryContext;

	ColumnCode();

	void initialize(
			ExprFactoryContext &cxt, const Expression *expr,
			const size_t *argIndex);

	CursorImpl *cursor_;
	ContainerColumn column_;
};

template<RowArrayType RAType, typename T, typename S, bool Nullable>
class SQLContainerImpl::ColumnEvaluator {
private:
	typedef SQLValues::ValueUtils ValueUtils;
	typedef SQLExprs::ExprContext ExprContext;

	typedef typename T::ValueType ValueType;
	typedef typename S::ValueType SourceType;

public:
	typedef ColumnEvaluator EvaluatorType;
	typedef ColumnCode CodeType;
	typedef std::pair<ValueType, bool> ResultType;

	static ColumnEvaluator evaluator(const CodeType &code) {
		return ColumnEvaluator(code);
	}

	static bool check(const ResultType &value) {
		return !Nullable || !value.second;
	}

	static const ValueType& get(const ResultType &value) {
		return value.first;
	}

	ResultType eval(ExprContext &cxt) const;

private:
	explicit ColumnEvaluator(const CodeType &code) : code_(code) {
	}

	const CodeType &code_;
};

class SQLContainerImpl::ColumnExpression {
private:
	typedef SQLExprs::ExprCode ExprCode;
	typedef SQLExprs::ExprFactoryContext ExprFactoryContext;

	typedef SQLCoreExprs::ColumnExpression CoreExpr;
	typedef SQLCoreExprs::VariantTypes VariantTypes;

	enum {
		COUNT_FULL_TYPE = CoreExpr::COUNT_FULL_TYPE,
		COUNT_NULLABLE = CoreExpr::COUNT_NULLABLE,
		COUNT_CONTAINER_SOURCE = CoreExpr::COUNT_CONTAINER_SOURCE,

		OFFSET_CONTAINER = CoreExpr::OFFSET_CONTAINER,
		OFFSET_END = CoreExpr::OFFSET_END
	};

public:
	template<size_t I>
	struct RowArrayTypeOf {
		static const BaseContainer::RowArrayType RA_TYPE = (I == 0 ?
				BaseContainer::ROW_ARRAY_GENERAL :
				BaseContainer::ROW_ARRAY_PLAIN);
	};

	struct VariantTraits {
		enum {
			TOTAL_VARIANT_COUNT = OFFSET_END,

			VARIANT_BEGIN = OFFSET_CONTAINER,
			VARIANT_END = OFFSET_END
		};

		static size_t resolveVariant(
				ExprFactoryContext &cxt, const ExprCode &code) {
			return CoreExpr::VariantTraits::resolveVariant(cxt, code);
		}
	};

	template<size_t V, bool = (V <= OFFSET_CONTAINER)>
	struct VariantTraitsBase {
		typedef SQLValues::Types::Any ValueTypeTag;
		typedef ValueTypeTag ResultTypeTag;

		typedef SQLCoreExprs::ConstEvaluator<ValueTypeTag> Evaluator;
		typedef typename Evaluator::CodeType Code;
		typedef util::TrueType Checking;
	};

	template<size_t V>
	struct VariantTraitsBase<V, false> {
		enum {
			SUB_OFFSET = (V - OFFSET_CONTAINER - 1),
			SUB_SOURCE =
					(SUB_OFFSET / (COUNT_NULLABLE * COUNT_FULL_TYPE)),
			SUB_NULLABLE =
					(((SUB_OFFSET / COUNT_FULL_TYPE) % COUNT_NULLABLE) != 0),
			SUB_TYPE = (SUB_OFFSET % COUNT_FULL_TYPE)
		};

		static const BaseContainer::RowArrayType SUB_RA_TYPE =
				RowArrayTypeOf<SUB_SOURCE>::RA_TYPE;

		typedef typename VariantTypes::ColumnTypeByFull<
				SUB_TYPE>::ValueTypeTag ValueTypeTag;
		typedef ValueTypeTag ResultTypeTag;

		typedef ColumnEvaluator<
				SUB_RA_TYPE, ResultTypeTag, ValueTypeTag,
				SUB_NULLABLE> Evaluator;
		typedef typename Evaluator::CodeType Code;
		typedef util::FalseType Checking;
	};

	template<size_t V>
	struct VariantAt {
		typedef CoreExpr::VariantBaseAt<V, ColumnExpression> VariantType;
	};
};

class SQLContainerImpl::IdExpression {
private:
	typedef SQLExprs::ExprCode ExprCode;
	typedef SQLExprs::ExprContext ExprContext;
	typedef SQLExprs::ExprFactoryContext ExprFactoryContext;

	typedef SQLCoreExprs::IdExpression CoreExpr;
	typedef SQLCoreExprs::VariantTypes VariantTypes;

public:
	struct VariantTraits {
		enum {
			TOTAL_VARIANT_COUNT = CoreExpr::OFFSET_END,

			VARIANT_BEGIN = CoreExpr::OFFSET_CONTAINER,
			VARIANT_END = CoreExpr::OFFSET_END
		};

		static size_t resolveVariant(
				ExprFactoryContext &cxt, const ExprCode &code) {
			return CoreExpr::VariantTraits::resolveVariant(cxt, code);
		}
	};

	class VariantBase : public SQLExprs::Expression {
	public:
		virtual ~VariantBase() = 0;
		void initializeCustom(ExprFactoryContext &cxt, const ExprCode &code);

	protected:
		ColumnCode columnCode_;
	};

	template<size_t V>
	class VariantAt : public VariantBase {
	public:
		typedef VariantAt VariantType;

		virtual TupleValue eval(ExprContext &cxt) const;

	private:
		enum {
			SUB_OFFSET = (V - CoreExpr::OFFSET_CONTAINER),
			SUB_SOURCE = SUB_OFFSET,
			SUB_NULLABLE = 0
		};

		static const BaseContainer::RowArrayType SUB_RA_TYPE =
				ColumnExpression::template RowArrayTypeOf<
						SUB_SOURCE>::RA_TYPE;

		typedef SQLValues::Types::Long ValueTypeTag;
		typedef ValueTypeTag ResultTypeTag;

		typedef ColumnEvaluator<
				SUB_RA_TYPE, ResultTypeTag, ValueTypeTag,
				SUB_NULLABLE> Evaluator;
	};
};

class SQLContainerImpl::ComparisonExpressionBase {
private:
	typedef SQLExprs::ExprCode ExprCode;
	typedef SQLExprs::ExprFactoryContext ExprFactoryContext;

	typedef SQLCoreExprs::ComparisonExpressionBase CoreExpr;
	typedef SQLCoreExprs::VariantTypes VariantTypes;

	enum {
		COUNT_NULLABLE = CoreExpr::COUNT_NULLABLE,
		COUNT_CONTAINER_SOURCE = CoreExpr::COUNT_CONTAINER_SOURCE,
		COUNT_BASIC_TYPE = CoreExpr::COUNT_BASIC_TYPE,

		OFFSET_CONTAINER = CoreExpr::OFFSET_CONTAINER,
		OFFSET_END = CoreExpr::OFFSET_END
	};

public:
	struct VariantTraits {
		enum {
			TOTAL_VARIANT_COUNT = OFFSET_END,

			VARIANT_BEGIN = OFFSET_CONTAINER,
			VARIANT_END = OFFSET_END
		};

		static size_t resolveVariant(
				ExprFactoryContext &cxt, const ExprCode &code) {
			return CoreExpr::VariantTraits::resolveVariant(cxt, code);
		}
	};

	template<size_t V>
	struct VariantTraitsBase {
		enum {
			SUB_OFFSET = V - OFFSET_CONTAINER,
			SUB_SOURCE =
					(SUB_OFFSET / (COUNT_NULLABLE * COUNT_BASIC_TYPE)),
			SUB_NULLABLE =
					(((SUB_OFFSET / COUNT_BASIC_TYPE) % COUNT_NULLABLE) != 0),
			SUB_TYPE = (SUB_OFFSET % COUNT_BASIC_TYPE)
		};

		static const BaseContainer::RowArrayType SUB_RA_TYPE =
				ColumnExpression::RowArrayTypeOf<SUB_SOURCE>::RA_TYPE;

		typedef typename VariantTypes::ColumnTypeByBasic<
				SUB_TYPE>::ValueTypeTag SourceType;
		typedef typename VariantTypes::ColumnTypeByBasic<
				SUB_TYPE>::CompTypeTag CompTypeTag;

		typedef ColumnEvaluator<
				SUB_RA_TYPE, CompTypeTag, SourceType, SUB_NULLABLE> Arg1;
		typedef SQLCoreExprs::ConstEvaluator<CompTypeTag> Arg2;

		typedef typename Arg1::CodeType Code1;
		typedef typename Arg2::CodeType Code2;

		typedef typename SQLValues::ValueBasicComparator::TypeAt<
				CompTypeTag> TypedComparator;

		typedef util::FalseType Checking;
	};
};

template<typename Func>
class SQLContainerImpl::ComparisonExpression {
	typedef ComparisonExpressionBase Base;
	typedef SQLCoreExprs::ComparisonExpressionBase CoreExpr;
public:
	typedef Base::VariantTraits VariantTraits;

	template<size_t V>
	struct VariantAt {
		typedef CoreExpr::VariantBaseAt<Func, V, Base> VariantType;
	};
};

class SQLContainerImpl::GeneralCondEvaluator {
private:
	typedef SQLExprs::Expression Expression;
	typedef SQLExprs::ExprFactoryContext ExprFactoryContext;

public:
	typedef const Expression &EvaluatorType;
	typedef GeneralCondEvaluator CodeType;
	typedef TupleValue ResultType;

	static EvaluatorType evaluator(const CodeType &code) {
		return *code.expr_;
	}

	static bool check(const ResultType &value) {
		return !ValueUtils::isNull(value);
	}

	static bool get(const ResultType &value) {
		return ValueUtils::getValue<bool>(value);
	}

	void initialize(
			ExprFactoryContext &cxt, const Expression *expr,
			const size_t *argIndex);

private:
	const Expression *expr_;
};

template<typename Func, size_t V>
class SQLContainerImpl::ComparisonEvaluator {
private:
	typedef SQLExprs::Expression Expression;
	typedef SQLExprs::ExprContext ExprContext;
	typedef SQLExprs::ExprFactoryContext ExprFactoryContext;

	enum {
		RESULT_NULLABLE =
				ComparisonExpressionBase::VariantTraitsBase<V>::SUB_NULLABLE
	};

public:
	typedef const ComparisonEvaluator &EvaluatorType;
	typedef ComparisonEvaluator CodeType;
	typedef typename util::BoolType<RESULT_NULLABLE>::Result NullableType;
	typedef typename util::Conditional<
			RESULT_NULLABLE, TupleValue, bool>::Type ResultType;

	static EvaluatorType evaluator(const CodeType &code) {
		return code;
	}

	static bool check(const bool&) {
		return true;
	}
	static bool check(const TupleValue &value) {
		return !ValueUtils::isNull(value);
	}

	static bool get(const bool &value) {
		return value;
	}
	static bool get(const TupleValue &value) {
		return ValueUtils::getValue<bool>(value);
	}

	ResultType eval(ExprContext &cxt) const {
		return toResult(expr_.evalAsBool(cxt), NullableType());
	}

	void initialize(
			ExprFactoryContext &cxt, const Expression *expr,
			const size_t *argIndex);

private:
	static TupleValue toResult(const TupleValue &src, const util::TrueType&) {
		return src;
	}
	static bool toResult(const TupleValue &src, const util::FalseType&) {
		return ValueUtils::getValue<bool>(src);
	}

	static TupleValue toResult(const bool &src, const util::TrueType&) {
		return TupleValue(src);
	}
	static bool toResult(const bool &src, const util::FalseType&) {
		return src;
	}

	typedef ComparisonExpression<Func> CompExprType;
	typedef typename CompExprType::template VariantAt<
			V>::VariantType CompExprVariantType;

	CompExprVariantType expr_;
};

class SQLContainerImpl::EmptyCondEvaluator {
private:
	typedef SQLExprs::Expression Expression;
	typedef SQLExprs::ExprContext ExprContext;
	typedef SQLExprs::ExprFactoryContext ExprFactoryContext;

public:
	typedef const EmptyCondEvaluator &EvaluatorType;
	typedef EmptyCondEvaluator CodeType;
	typedef bool ResultType;

	static EvaluatorType evaluator(const CodeType &code) {
		return code;
	}

	static bool check(const ResultType &value) {
		return true;
	}

	static bool get(const ResultType &value) {
		static_cast<void>(value);
		return true;
	}

	ResultType eval(ExprContext &cxt) const {
		static_cast<void>(cxt);
		return true;
	}

	void initialize(
			ExprFactoryContext &cxt, const Expression *expr,
			const size_t *argIndex);
};

class SQLContainerImpl::GeneralProjector {
private:
	typedef SQLOps::ProjectionFactoryContext ProjectionFactoryContext;

public:
	GeneralProjector() : proj_(NULL) {
	}

	void project(OpContext &cxt) {
		proj_->project(cxt);
	}

	UpdatorEntry getUpdator() {
		return UpdatorEntry(&updateFunc, this);
	}

	static void updateFunc(OpContext &cxt, void *src) {
		static_cast<GeneralProjector*>(src)->update(cxt);
	}

	bool update(OpContext &cxt);
	void initialize(ProjectionFactoryContext &cxt, const Projection *proj);

private:
	const Projection *proj_;
};

class SQLContainerImpl::CountProjector {
private:
	typedef SQLOps::ProjectionFactoryContext ProjectionFactoryContext;

public:
	CountProjector() : counter_(0) {
	}

	void project(OpContext &cxt) {
		++counter_;
	}

	UpdatorEntry getUpdator() {
		return UpdatorEntry(&updateFunc, this);
	}

	static void updateFunc(OpContext &cxt, void *src) {
		static_cast<CountProjector*>(src)->update(cxt);
	}

	bool update(OpContext &cxt);
	void initialize(ProjectionFactoryContext &cxt, const Projection *proj);

private:
	int64_t counter_;
	SQLValues::SummaryColumn column_;
};

class SQLContainerImpl::CursorRef {
private:
	typedef SQLExprs::ExprContext ExprContext;

public:
	explicit CursorRef(util::StackAllocator &alloc);

	CursorImpl& resolveCursor();
	void setCursor(CursorImpl *cursor);

	void addContextRef(OpContext **ref);
	void addContextRef(ExprContext **ref);

	void updateAll(OpContext &cxt);

private:
	typedef util::Vector<OpContext**> ContextRefList;
	typedef util::Vector<ExprContext**> ExprContextRefList;

	CursorImpl *cursor_;

	ContextRefList cxtRefList_;
	ExprContextRefList exprCxtRefList_;
};

template<BaseContainer::RowArrayType RAType>
struct SQLContainerImpl::ScannerHandler {
	static const RowArrayType ROW_ARRAY_TYPE = RAType;

	ScannerHandler(OpContext &cxt, CursorImpl &cursor, const Projection &proj);

	template<BaseContainer::RowArrayType InRAType>
	void operator()(
			TransactionContext &txn,
			typename BaseContainer::RowArrayImpl<
					BaseContainer, InRAType> &rowArrayImpl);

	bool update(TransactionContext &txn);

	OpContext &cxt_;
	CursorImpl &cursor_;
	const Projection &proj_;
};

class SQLContainerImpl::ScannerHandlerBase {
private:
	typedef SQLExprs::ExprType ExprType;
	typedef SQLExprs::ExprContext ExprContext;

	typedef SQLCoreExprs::ComparisonExpressionBase CoreCompExpr;

	typedef SQLOps::ProjectionFactoryContext ProjectionFactoryContext;

	enum {
		COUNT_NULLABLE = CoreCompExpr::COUNT_NULLABLE,
		COUNT_BASIC_TYPE = CoreCompExpr::COUNT_BASIC_TYPE,
		COUNT_CONTAINER_SOURCE = CoreCompExpr::COUNT_CONTAINER_SOURCE,
		COUNT_COMP_BASE = COUNT_NULLABLE * COUNT_BASIC_TYPE,

		OFFSET_COMP_CONTAINER = CoreCompExpr::OFFSET_CONTAINER,

		EMPTY_FILTER_NUM = 1,
		COUNT_NO_COMP_FILTER = 2,

		COUNT_FUNC = 6,
		COUNT_FILTER = COUNT_NO_COMP_FILTER + COUNT_FUNC * COUNT_COMP_BASE,
		COUNT_PROJECTION = 2
	};

public:
	typedef std::pair<
				ProjectionFactoryContext*,
				ContainerRowScanner::HandlerSet*> FactoryContext;

	struct HandlerVariantUtils;

	template<size_t I> struct CompFuncOf;
	template<size_t V> struct VariantTraitsBase;

	struct VariantTraits;
	template<size_t V> struct VariantAt;
};

struct SQLContainerImpl::ScannerHandlerBase::HandlerVariantUtils {
private:
	typedef SQLExprs::ExprCode ExprCode;
	typedef ExprCode::InputSourceType InputSourceType;

public:
	static size_t toInputSourceNumber(InputSourceType type);
	static size_t toCompFuncNumber(ExprType type);

	static void setUpContextRef(
			ProjectionFactoryContext &cxt,
			OpContext **ref, ExprContext **exprRef);
	static void addUpdator(
			ProjectionFactoryContext &cxt, const UpdatorEntry &updator);

	static std::pair<const Expression*, const Projection*>
	getFilterProjectionElements(const Projection &src);
};

template<size_t I>
struct SQLContainerImpl::ScannerHandlerBase::CompFuncOf {
	typedef SQLCoreExprs::Functions Functions;

	typedef
			typename util::Conditional<(I == 0), Functions::Lt,
			typename util::Conditional<(I == 1), Functions::Gt,
			typename util::Conditional<(I == 2), Functions::Le,
			typename util::Conditional<(I == 3), Functions::Ge,
			typename util::Conditional<(I == 4), Functions::Eq,
			typename util::Conditional<(I == 5), Functions::Ne,
			void>::Type>::Type>::Type>::Type>::Type>::Type FuncType;
};

template<size_t V>
struct SQLContainerImpl::ScannerHandlerBase::VariantTraitsBase {
	enum {
		SUB_SOURCE = (V / (COUNT_PROJECTION * COUNT_FILTER)),
		SUB_PROJECTION = ((V / COUNT_FILTER) % COUNT_PROJECTION),
		SUB_FILTER = (V % COUNT_FILTER),

		SUB_EMPTY_FILTERING = (SUB_FILTER == EMPTY_FILTER_NUM),
		SUB_COMP_FILTERING = (SUB_FILTER >= COUNT_NO_COMP_FILTER),

		SUB_COMP = (SUB_COMP_FILTERING ?
				SUB_FILTER - COUNT_NO_COMP_FILTER : 0),
		SUB_COMP_VARIANT = (OFFSET_COMP_CONTAINER +
				SUB_SOURCE * COUNT_COMP_BASE +
				SUB_COMP % COUNT_COMP_BASE),
		SUB_COMP_FUNC = (SUB_COMP / COUNT_COMP_BASE)
	};

	typedef typename CompFuncOf<SUB_COMP_FUNC>::FuncType CompFuncType;

	typedef ComparisonEvaluator<
			CompFuncType, SUB_COMP_VARIANT> CompEvaluatorType;

	typedef
			typename util::Conditional<
					SUB_EMPTY_FILTERING, EmptyCondEvaluator,
			typename util::Conditional<
					SUB_COMP_FILTERING, CompEvaluatorType,
					GeneralCondEvaluator>::Type>::Type EvaluatorType;

	typedef typename util::Conditional<
			(SUB_PROJECTION == 0),
			GeneralProjector, CountProjector>::Type ProjectorType;
};

struct SQLContainerImpl::ScannerHandlerBase::VariantTraits {
	enum {
		TOTAL_VARIANT_COUNT =
				COUNT_CONTAINER_SOURCE * COUNT_PROJECTION * COUNT_FILTER,

		VARIANT_BEGIN = 0,
		VARIANT_END = TOTAL_VARIANT_COUNT
	};

	static size_t resolveVariant(FactoryContext &cxt, const Projection &proj);
};

template<size_t V>
struct SQLContainerImpl::ScannerHandlerBase::VariantAt {
private:
	typedef VariantTraitsBase<V> TraitsBase;

public:
	typedef VariantAt VariantType;

	static const BaseContainer::RowArrayType ROW_ARRAY_TYPE =
			ColumnExpression::RowArrayTypeOf<TraitsBase::SUB_SOURCE>::RA_TYPE;

	VariantAt(ProjectionFactoryContext &cxt, const Projection &proj);

	template<BaseContainer::RowArrayType InRAType>
	void operator()(
			TransactionContext&,
			typename BaseContainer::RowArrayImpl<
					BaseContainer, InRAType>&);

	bool update(TransactionContext&);

private:
	typedef typename TraitsBase::EvaluatorType Evaluator;
	typedef typename TraitsBase::ProjectorType Projector;

	typedef typename Evaluator::CodeType EvaluatorCode;
	typedef typename Evaluator::ResultType ResultType;

	OpContext *cxt_;
	ExprContext *exprCxt_;

	EvaluatorCode condCode_;
	Projector projector_;
};

class SQLContainerImpl::ScannerHandlerFactory {
public:
	typedef ScannerHandlerBase::FactoryContext FactoryContext;

	typedef ContainerRowScanner::HandlerSet& (*FactoryFunc)(
				FactoryContext&, const Projection&);

	static const ScannerHandlerFactory& getInstance();
	static ScannerHandlerFactory& getInstanceForRegistrar();

	ContainerRowScanner& create(
			OpContext &cxt, CursorImpl &cursor, RowArray &rowArray,
			const Projection &proj) const;

	ContainerRowScanner::HandlerSet& create(
				FactoryContext &cxt, const Projection &proj) const;

	void setFactoryFunc(FactoryFunc func);

private:
	ScannerHandlerFactory();

	ScannerHandlerFactory(const ScannerHandlerFactory&);
	ScannerHandlerFactory& operator=(const ScannerHandlerFactory&);

	static const ScannerHandlerFactory &FACTORY_INSTANCE;

	FactoryFunc factoryFunc_;
};

class SQLContainerImpl::BaseRegistrar {
public:
	typedef SQLExprs::ExprRegistrar::VariantTraits ExprVariantTraits;
	struct ScannerVariantTraits;

	explicit BaseRegistrar(const BaseRegistrar &subRegistrar) throw();

	virtual void operator()() const;

protected:
	BaseRegistrar() throw();

	template<SQLExprs::ExprType T, typename E>
	void addExprVariants() const;

	template<typename S>
	void addScannerVariants() const;
	template<typename S, size_t Begin, size_t End>
	void addScannerVariants() const;

private:
	ScannerHandlerFactory *factory_;
};

struct SQLContainerImpl::BaseRegistrar::ScannerVariantTraits {
	typedef ScannerHandlerFactory::FactoryContext ContextType;
	typedef Projection CodeType;
	typedef ContainerRowScanner::HandlerSet ObjectType;

	template<typename V>
	static ObjectType& create(ContextType &cxt, const CodeType &code);
};

class SQLContainerImpl::DefaultRegistrar :
		public SQLContainerImpl::BaseRegistrar {
public:
	virtual void operator()() const;

private:
	static const BaseRegistrar REGISTRAR_INSTANCE;
};

struct SQLContainerImpl::SearchResult {
	explicit SearchResult(util::StackAllocator &alloc);

	util::XArray<OId> oIdList_;
	util::XArray<OId> mvccOIdList_;
	bool asRowArray_;
};

class SQLContainerImpl::CursorScope {
public:
	CursorScope(OpContext &cxt, CursorImpl &cursor);
	~CursorScope();

	TransactionContext& getTransactionContext();
	ResultSet& getResultSet();

private:
	void initialize(OpContext &cxt);
	void clear(bool force);
	void clearResultSet(bool force);

	static const DataStore::Latch* getStoreLatch(
			TransactionContext &txn, OpContext &cxt,
			util::LocalUniquePtr<DataStore::Latch> &latch);
	static BaseContainer* getContainer(
			TransactionContext &txn, OpContext &cxt,
			const DataStore::Latch *latch, const ContainerLocation &location,
			bool expired, ContainerType type);
	static BaseContainer* checkContainer(
			BaseContainer *container, const ContainerLocation &location,
			SchemaVersionId schemaVersionId);
	static ResultSet* prepareResultSet(
			OpContext &cxt, TransactionContext &txn, BaseContainer *container,
			util::LocalUniquePtr<ResultSetGuard> &rsGuard,
			const CursorImpl &cursor);
	static ResultSetId findResultSetId(ResultSet *resultSet);

	CursorImpl &cursor_;
	TransactionContext &txn_;
	TransactionService *txnSvc_;
	util::LocalUniquePtr<DataStore::Latch> latch_;
	StackAllocAutoPtr<BaseContainer> containerAutoPtr_;
	util::LocalUniquePtr<ResultSetGuard> rsGuard_;
	ResultSet *resultSet_;
};

class SQLContainerImpl::MetaScanHandler : public MetaProcessor::RowHandler {
public:
	MetaScanHandler(
			OpContext &cxt, CursorImpl &cursor, const Projection &proj);

	virtual void operator()(
			TransactionContext &txn,
			const MetaProcessor::ValueList &valueList);

private:
	OpContext &cxt_;
	CursorImpl &cursor_;
	const Projection &proj_;
};

class SQLContainerImpl::TupleUpdateRowIdHandler :
		public ResultSet::UpdateRowIdHandler {
public:
	TupleUpdateRowIdHandler(OpContext &cxt, RowId maxRowId);
	~TupleUpdateRowIdHandler();

	void destroy() throw();

	void close();
	void operator()(RowId rowId, uint64_t id, ResultSet::UpdateOperation op);

	bool getRowIdList(util::XArray<RowId> &list, size_t limit);

	TupleUpdateRowIdHandler* share(util::StackAllocator &alloc);

private:
	enum {
		ID_STORE_COLUMN_COUNT = 1
	};

	typedef TupleList::Column IdStoreColumnList[ID_STORE_COLUMN_COUNT];

	struct Shared {
		Shared(OpContext &cxt, RowId maxRowId);

		size_t refCount_;
		util::LocalUniquePtr<OpContext> cxt_;

		IdStoreColumnList columnList_;
		uint32_t extraStoreIndex_;
		uint32_t rowIdStoreIndex_;

		RowId maxRowId_;
	};

	explicit TupleUpdateRowIdHandler(Shared &shared);

	static uint32_t createRowIdStore(
			util::LocalUniquePtr<OpContext> &cxt, const OpContext::Source &cxtSrc,
			IdStoreColumnList &columnList, uint32_t &extraStoreIndex);

	static void copyRowIdStore(
			OpContext &cxt, uint32_t srcIndex, uint32_t destIndex,
			const IdStoreColumnList &columnList);

	static util::ObjectPool<Shared, util::Mutex> sharedPool_;

	Shared *shared_;
};

struct SQLContainerImpl::ContainerValueUtils {
	static void toContainerValue(
			util::StackAllocator &alloc, const TupleValue &src, Value &dest);
	static bool toTupleValue(
			util::StackAllocator &alloc, const Value &src, TupleValue &dest,
			bool failOnUnsupported);

	static TupleColumnType toTupleColumnType(
			ContainerColumnType src, bool nullable);

	static TupleColumnType toRawTupleColumnType(ContainerColumnType src);

	static bool toContainerColumnType(
			TupleColumnType src, ContainerColumnType &dest, bool failOnError);
};

struct SQLContainerImpl::TQLTool {
	static SyntaxExpr genPredExpr(
			util::StackAllocator &alloc, const Query &query);
	static SyntaxExpr genPredExpr(
			util::StackAllocator &alloc, const BoolExpr *src, bool *unresolved);
	static SyntaxExpr genPredExpr(
			util::StackAllocator &alloc, const Expr *src, bool *unresolved);

	static SyntaxExpr filterPredExpr(
			util::StackAllocator &alloc, const SyntaxExpr &src, bool *unresolved);

	static SyntaxExpr genUnresolvedTopExpr(util::StackAllocator &alloc);
	static SyntaxExpr genUnresolvedLeafExpr(
			util::StackAllocator &alloc, bool *unresolved);

	static SyntaxExpr genConstExpr(
			util::StackAllocator &alloc, const TupleValue &value);

	template<typename T>
	static SyntaxExprRefList* genExprList(
			util::StackAllocator &alloc, const util::XArray<T*> *src,
			bool *unresolved, size_t argCount);
	template<typename T>
	static SyntaxExprRefList* genExprList(
			util::StackAllocator &alloc, const util::XArray<T*> *src,
			bool *unresolved, size_t minArgCount, size_t maxArgCount);
};

#endif
